from py42.__version__ import __version__ as py42version

from code42cli.__version__ import __version__ as cliversion


PRODUCT_NAME = "code42cli"
MAIN_COMMAND = "code42"
BANNER = f"""\b
 dP""b8  dP"Yb  8888b. 888888  dP88  oP"Yb.
dP   `" dP   Yb 8I  Yb 88__   dP 88  "' dP'
Yb      Yb   dP 8I  dY 88""  d888888   dP'
 YboodP  YbodP  8888Y" 888888    88  .d8888

code42cli version {cliversion}, by Code42 Software.
powered by py42 version {py42version}."""
